@@ -38,12 +38,12 @@
 #                                                   'volume_id': '1234'})
 #             return original(manager, name_or_id, **kwargs)
 
+from unittest import mock
+from urllib import parse
+
 import ddt
 import fixtures
-import mock
 from requests_mock.contrib import fixture as requests_mock_fixture
-import six
-from six.moves.urllib import parse
 
 import cinderclient
 from cinderclient import api_versions
@@ -51,15 +51,14 @@ from cinderclient import base
 from cinderclient import client
 from cinderclient import exceptions
 from cinderclient import shell
+from cinderclient.tests.unit.fixture_data import keystone_client
+from cinderclient.tests.unit import utils
+from cinderclient.tests.unit.v3 import fakes
 from cinderclient import utils as cinderclient_utils
 from cinderclient.v3 import attachments
 from cinderclient.v3 import volume_snapshots
 from cinderclient.v3 import volumes
 
-from cinderclient.tests.unit.fixture_data import keystone_client
-from cinderclient.tests.unit import utils
-from cinderclient.tests.unit.v3 import fakes
-
 
 @ddt.ddt
 @mock.patch.object(client, 'Client', fakes.FakeClient)
@@ -321,7 +320,7 @@ class ShellTest(utils.TestCase):
         self.assert_call_contained(
             parse.urlencode(
                 {'extra_specs':
-                    {six.text_type('key'): six.text_type('value')}}))
+                    {'key': 'value'}}))
         self.assert_call_contained(parse.urlencode({'is_public': None}))
 
     def test_type_list_public(self):
@@ -389,6 +388,25 @@ class ShellTest(utils.TestCase):
               {'cmd': 'abc 1233',
                'body': {'instance_uuid': '1233',
                         'connector': {},
+                        'volume_uuid': '1234'}},
+              {'cmd': '1234',
+               'body': {'connector': {},
+                        'volume_uuid': '1234'}},
+              {'cmd': '1234 '
+                      '--connect True '
+                      '--ip 10.23.12.23 --host server01 '
+                      '--platform x86_xx '
+                      '--ostype 123 '
+                      '--multipath true '
+                      '--mountpoint /123 '
+                      '--initiator aabbccdd',
+               'body': {'connector': {'ip': '10.23.12.23',
+                                      'host': 'server01',
+                                      'os_type': '123',
+                                      'multipath': 'true',
+                                      'mountpoint': '/123',
+                                      'initiator': 'aabbccdd',
+                                      'platform': 'x86_xx'},
                         'volume_uuid': '1234'}})
     @mock.patch('cinderclient.utils.find_resource')
     @ddt.unpack
@@ -430,6 +448,27 @@ class ShellTest(utils.TestCase):
                'body': {'instance_uuid': '1233',
                         'connector': {},
                         'volume_uuid': '1234',
+                        'mode': 'ro'}},
+              {'cmd': '1234',
+               'body': {'connector': {},
+                        'volume_uuid': '1234',
+                        'mode': 'ro'}},
+              {'cmd': '1234 '
+                      '--connect True '
+                      '--ip 10.23.12.23 --host server01 '
+                      '--platform x86_xx '
+                      '--ostype 123 '
+                      '--multipath true '
+                      '--mountpoint /123 '
+                      '--initiator aabbccdd',
+               'body': {'connector': {'ip': '10.23.12.23',
+                                      'host': 'server01',
+                                      'os_type': '123',
+                                      'multipath': 'true',
+                                      'mountpoint': '/123',
+                                      'initiator': 'aabbccdd',
+                                      'platform': 'x86_xx'},
+                        'volume_uuid': '1234',
                         'mode': 'ro'}})
     @mock.patch('cinderclient.utils.find_resource')
     @ddt.unpack
@@ -1149,11 +1188,7 @@ class ShellTest(utils.TestCase):
                           self.run_command, '--os-volume-api-version 3.32 '
                           'service-set-log')
         set_levels_mock.assert_not_called()
-        # Different error message from argparse library in Python 2 and 3
-        if six.PY3:
-            msg = 'the following arguments are required: <log-level>'
-        else:
-            msg = 'too few arguments'
+        msg = 'the following arguments are required: <log-level>'
         error_mock.assert_called_once_with(msg)
 
     @ddt.data('debug', 'DEBUG', 'info', 'INFO', 'warning', 'WARNING', 'error',
@@ -1610,3 +1645,190 @@ class ShellTest(utils.TestCase):
     def test_transfer_list_with_filters(self, command, expected):
         self.run_command('--os-volume-api-version 3.52 %s' % command)
         self.assert_called('GET', expected)
+
+    def test_default_type_set(self):
+        self.run_command('--os-volume-api-version 3.62 default-type-set '
+                         '4c298f16-e339-4c80-b934-6cbfcb7525a0 '
+                         '629632e7-99d2-4c40-9ae3-106fa3b1c9b7')
+        body = {
+            'default_type':
+                {
+                    'volume_type': '4c298f16-e339-4c80-b934-6cbfcb7525a0'
+                }
+            }
+        self.assert_called(
+            'PUT', 'v3/default-types/629632e7-99d2-4c40-9ae3-106fa3b1c9b7',
+            body=body)
+
+    def test_default_type_list_project(self):
+        self.run_command('--os-volume-api-version 3.62 default-type-list '
+                         '--project-id 629632e7-99d2-4c40-9ae3-106fa3b1c9b7')
+        self.assert_called(
+            'GET', 'v3/default-types/629632e7-99d2-4c40-9ae3-106fa3b1c9b7')
+
+    def test_default_type_list(self):
+        self.run_command('--os-volume-api-version 3.62 default-type-list')
+        self.assert_called('GET', 'v3/default-types')
+
+    def test_default_type_delete(self):
+        self.run_command('--os-volume-api-version 3.62 default-type-unset '
+                         '629632e7-99d2-4c40-9ae3-106fa3b1c9b7')
+        self.assert_called(
+            'DELETE', 'v3/default-types/629632e7-99d2-4c40-9ae3-106fa3b1c9b7')
+
+    def test_restore(self):
+        self.run_command('backup-restore 1234')
+        self.assert_called('POST', '/backups/1234/restore')
+
+    def test_restore_with_name(self):
+        self.run_command('backup-restore 1234 --name restore_vol')
+        expected = {'restore': {'volume_id': None, 'name': 'restore_vol'}}
+        self.assert_called('POST', '/backups/1234/restore',
+                           body=expected)
+
+    def test_restore_with_name_error(self):
+        self.assertRaises(exceptions.CommandError, self.run_command,
+                          'backup-restore 1234 --volume fake_vol --name '
+                          'restore_vol')
+
+    def test_restore_with_az(self):
+        self.run_command('--os-volume-api-version 3.47 backup-restore 1234 '
+                         '--name restore_vol --availability-zone restore_az')
+        expected = {'volume': {'size': 10,
+                               'name': 'restore_vol',
+                               'availability_zone': 'restore_az',
+                               'backup_id': '1234',
+                               'metadata': {},
+                               'imageRef': None,
+                               'source_volid': None,
+                               'consistencygroup_id': None,
+                               'snapshot_id': None,
+                               'volume_type': None,
+                               'description': None}}
+        self.assert_called('POST', '/volumes', body=expected)
+
+    def test_restore_with_az_microversion_error(self):
+        self.assertRaises(exceptions.UnsupportedAttribute, self.run_command,
+                          '--os-volume-api-version 3.46 backup-restore 1234 '
+                          '--name restore_vol --availability-zone restore_az')
+
+    def test_restore_with_volume_type(self):
+        self.run_command('--os-volume-api-version 3.47 backup-restore 1234 '
+                         '--name restore_vol --volume-type restore_type')
+        expected = {'volume': {'size': 10,
+                               'name': 'restore_vol',
+                               'volume_type': 'restore_type',
+                               'backup_id': '1234',
+                               'metadata': {},
+                               'imageRef': None,
+                               'source_volid': None,
+                               'consistencygroup_id': None,
+                               'snapshot_id': None,
+                               'availability_zone': None,
+                               'description': None}}
+        self.assert_called('POST', '/volumes', body=expected)
+
+    def test_restore_with_volume_type_microversion_error(self):
+        self.assertRaises(exceptions.UnsupportedAttribute, self.run_command,
+                          '--os-volume-api-version 3.46 backup-restore 1234 '
+                          '--name restore_vol --volume-type restore_type')
+
+    def test_restore_with_volume_type_and_az_no_name(self):
+        self.run_command('--os-volume-api-version 3.47 backup-restore 1234 '
+                         '--volume-type restore_type '
+                         '--availability-zone restore_az')
+        expected = {'volume': {'size': 10,
+                               'name': 'restore_backup_1234',
+                               'volume_type': 'restore_type',
+                               'availability_zone': 'restore_az',
+                               'backup_id': '1234',
+                               'metadata': {},
+                               'imageRef': None,
+                               'source_volid': None,
+                               'consistencygroup_id': None,
+                               'snapshot_id': None,
+                               'description': None}}
+        self.assert_called('POST', '/volumes', body=expected)
+
+    @ddt.data(
+        {
+            'volume': '1234',
+            'name': None,
+            'volume_type': None,
+            'availability_zone': None,
+        }, {
+            'volume': '1234',
+            'name': 'ignored',
+            'volume_type': None,
+            'availability_zone': None,
+        }, {
+            'volume': None,
+            'name': 'sample-volume',
+            'volume_type': 'sample-type',
+            'availability_zone': None,
+        }, {
+            'volume': None,
+            'name': 'sample-volume',
+            'volume_type': None,
+            'availability_zone': 'az1',
+        }, {
+            'volume': None,
+            'name': 'sample-volume',
+            'volume_type': None,
+            'availability_zone': 'different-az',
+        }, {
+            'volume': None,
+            'name': None,
+            'volume_type': None,
+            'availability_zone': 'different-az',
+        },
+    )
+    @ddt.unpack
+    @mock.patch('cinderclient.utils.print_dict')
+    @mock.patch('cinderclient.tests.unit.v3.fakes_base._stub_restore')
+    def test_do_backup_restore(self,
+                               mock_stub_restore,
+                               mock_print_dict,
+                               volume,
+                               name,
+                               volume_type,
+                               availability_zone):
+
+        # Restore from the fake '1234' backup.
+        cmd = '--os-volume-api-version 3.47 backup-restore 1234'
+
+        if volume:
+            cmd += ' --volume %s' % volume
+        if name:
+            cmd += ' --name %s' % name
+        if volume_type:
+            cmd += ' --volume-type %s' % volume_type
+        if availability_zone:
+            cmd += ' --availability-zone %s' % availability_zone
+
+        if name or volume:
+            volume_name = 'sample-volume'
+        else:
+            volume_name = 'restore_backup_1234'
+
+        mock_stub_restore.return_value = {'volume_id': '1234',
+                                          'volume_name': volume_name}
+
+        self.run_command(cmd)
+
+        # Check whether mock_stub_restore was called in order to determine
+        # whether the restore command invoked the backup-restore API. If
+        # mock_stub_restore was not called then this indicates the command
+        # invoked the volume-create API to restore the backup to a new volume
+        # of a specific volume type, or in a different AZ (the fake '1234'
+        # backup is in az1).
+        if volume_type or availability_zone == 'different-az':
+            mock_stub_restore.assert_not_called()
+        else:
+            mock_stub_restore.assert_called_once()
+
+        mock_print_dict.assert_called_once_with({
+            'backup_id': '1234',
+            'volume_id': '1234',
+            'volume_name': volume_name,
+            })
